# coding: utf-8
# 2019/11/20 @ tongshiwei


__all__ = ["extract_relations", "build_json_sequence"]

from longling import path_append
from .junyi import build_knowledge_graph
from .KnowledgeTracing import select_n_most_frequent_students


def extract_relations(src_root: str = "../raw_data/junyi/", tar_root: str = "../data/junyi/data/"):
    build_knowledge_graph(
        src_root, tar_root,
        ku_dict_path="graph_vertex.json",
        prerequisite_path="prerequisite.json",
        similarity_path="similarity.json",
        difficulty_path="difficulty.json",
    )


def build_json_sequence(src_root: str = "../raw_data/junyi/", tar_root: str = "../data/junyi/data/",
                        ku_dict_path: str = "../data/junyi/data/graph_vertex.json", n: int = 1000):
    select_n_most_frequent_students(
        path_append(src_root, "junyi_ProblemLog_for_PSLC.txt", to_str=True),
        path_append(tar_root, "student_log_kt_", to_str=True),
        ku_dict_path,
        n,
    )



# coding: utf-8
# 2019/11/14 @ tongshiwei

from longling import path_append
from EduData.DataSet.junyi import extract_relations, build_json_sequence
from EduData.Task.KnowledgeTracing.format import tl2json, json2tl
from EduData.Task.KnowledgeTracing.statistics import analysis_records, analysis_edges
from EduData.Task.KnowledgeTracing.graph import dense_graph, transition_graph, correct_transition_graph
from EduData.Task.KnowledgeTracing.graph import similarity_graph


def test_junyi(shared_data_dir):
    src_root = path_append(shared_data_dir, "junyi", to_str=True)
    extract_relations(src_root, path_append(src_root, "data"))


def test_junyi_kt(shared_data_dir):
    src_root = path_append(shared_data_dir, "junyi", to_str=True)
    ku_dict_path = path_append(shared_data_dir, "junyi", "data", "graph_vertex.json")
    build_json_sequence(src_root, path_append(src_root, "data", to_str=True), ku_dict_path)


def test_json2tl(shared_data_dir):
    src = path_append(shared_data_dir, "junyi", "data", "student_log_kt_1000", to_str=True)
    tl_tar = path_append(shared_data_dir, "junyi", "data", "student_log_kt_1000.tl", to_str=True)
    json_tar = path_append(shared_data_dir, "junyi", "data", "student_log_kt_1000.json", to_str=True)
    json2tl(src, tl_tar)
    tl2json(tl_tar, json_tar, to_int=False)
    tl2json(tl_tar, json_tar, to_int=True)


def test_graph(shared_data_dir):
    json_src = path_append(shared_data_dir, "junyi", "data", "student_log_kt_1000.json", to_str=True)

    dense_graph(835, path_append(shared_data_dir, "dense_graph", to_str=True))
    trans_graph = path_append(shared_data_dir, "transition_graph", to_str=True)
    transition_graph(835, json_src, tar=trans_graph)
    ctrans_graph = path_append(shared_data_dir, "correct_transition_graph", to_str=True)
    correct_transition_graph(835, json_src, tar=ctrans_graph)

    ctrans_sim = path_append(shared_data_dir, "correct_transition_sim_graph", to_str=True)
    similarity_graph(835, ctrans_graph, ctrans_sim)


def test_analysis(shared_data_dir):
    src = path_append(shared_data_dir, "junyi", "data", "student_log_kt_1000", to_str=True)
    analysis_records(src)

    graph_src = path_append(shared_data_dir, "dense_graph", to_str=True)
    analysis_edges(graph_src)

    graph_src = path_append(shared_data_dir, "transition_graph", to_str=True)
    analysis_edges(graph_src, threshold=0.5)
    analysis_edges(graph_src, threshold=None)



# coding: utf-8
# 2019/11/14 @ tongshiwei

from EduData import get_data
from longling import path_append
import functools
import pytest

test_url_dict = {
    "tests":
        "http://base.ustc.edu.cn/data/tests/",
    "junyi":
        "http://base.ustc.edu.cn/data/tests/junyi/",
    "synthetic":
        "http://base.ustc.edu.cn/data/tests/synthetic",
    "ednet":
        "http://base.ustc.edu.cn/data/tests/EdNet/"
}

get_data = functools.partial(get_data, url_dict=test_url_dict)


@pytest.fixture(scope="session")
def shared_data_dir(tmp_path_factory):
    tmpdir = tmp_path_factory.mktemp("data")
    try:
        return path_append(get_data("tests", tmpdir, override=True), "tests")
    except Exception as e:
        raise e



# coding: utf-8
# create by tongshiwei on 2019/7/2

__all__ = ["URL_DICT", "get_data", "list_resources"]

import os
import re
from urllib.request import urlretrieve

import requests
from bs4 import BeautifulSoup
from longling import config_logging, LogLevel, path_append

try:
    from .utils import decompress, reporthook4urlretrieve
except (SystemError, ModuleNotFoundError):  # pragma: no cover
    from utils import decompress, reporthook4urlretrieve

DEFAULT_DATADIR = path_append("./", "", to_str=True)

logger = config_logging(logger="downloader", console_log_level=LogLevel.INFO)

prefix = 'http://base.ustc.edu.cn/data/'

URL_DICT = {
    "assistment-2009-2010-skill":
        "http://base.ustc.edu.cn/data/ASSISTment/2009_skill_builder_data_corrected.zip",
    "assistment-2012-2013-non-skill":
        "http://base.ustc.edu.cn/data/ASSISTment/2012-2013-data-with-predictions-4-final.zip",
    "assistment-2015":
        "http://base.ustc.edu.cn/data/ASSISTment/2015_100_skill_builders_main_problems.zip",
    "junyi":
        "http://base.ustc.edu.cn/data/JunyiAcademy_Math_Practicing_Log/junyi.rar",
    "KDD-CUP-2010":
        "http://base.ustc.edu.cn/data/KDD_Cup_2010/",
    "slepemapy.cz":
        "http://base.ustc.edu.cn/data/slepemapy.cz/",
    "synthetic":
        "http://base.ustc.edu.cn/data/synthetic/",
    "psychometrics":
        "http://base.ustc.edu.cn/data/psychometrics/",
    "psy":
        "http://base.ustc.edu.cn/data/psychometrics/",
    "ktbd":
        "http://base.ustc.edu.cn/data/ktbd/",
    "ktbd-a0910":
        "http://base.ustc.edu.cn/data/ktbd/assistment_2009_2010/",
    "ktbd-junyi":
        "http://base.ustc.edu.cn/data/ktbd/junyi/",
    "ktbd-synthetic":
        "http://base.ustc.edu.cn/data/ktbd/synthetic/",
    "ktbd-a0910c":
        "http://base.ustc.edu.cn/data/ktbd/a0910c/",
    "cdbd":
        "http://base.ustc.edu.cn/data/cdbd/",
    "cdbd-lsat":
        "http://base.ustc.edu.cn/data/cdbd/LSAT/",
    "cdbd-a0910":
        "http://base.ustc.edu.cn/data/cdbd/a0910/",
    "math2015":
        "http://staff.ustc.edu.cn/~qiliuql/data/math2015.rar",
    "ednet":
        "http://base.ustc.edu.cn/data/EdNet/",
    "ktbd-ednet":
        "http://base.ustc.edu.cn/data/ktbd/EdNet/"
}


def get_dataset_name():  # pragma: no cover
    urls = []
    for i in URL_DICT.values():
        if i not in urls:
            urls.append(i)
    url = prefix
    r = requests.get(url, timeout=30)
    r.raise_for_status()
    r.encoding = r.apparent_encoding
    soup = BeautifulSoup(r.text, "lxml")
    al = soup.find_all('a')
    for a in al:
        # 获得各个数据集名称
        h = a.get('href')
        if h[0] != '.':
            temp = prefix + h
            # 避免重复
            if temp not in urls:
                urls.append(temp)
                # 避免ASSISTment和junyi的重复
                if temp not in ['http://base.ustc.edu.cn/data/ASSISTment/',
                                'http://base.ustc.edu.cn/data/JunyiAcademy_Math_Practicing_Log/']:
                    URL_DICT[h[:-1]] = temp


def download_file(url, save_path, override):
    if os.path.exists(save_path) and override:
        logger.info(save_path + ' will be overridden.')
    elif os.path.exists(save_path):
        raise FileExistsError()

    logger.info(url + ' is saved as ' + save_path)
    urlretrieve(url, save_path, reporthook=reporthook4urlretrieve)
    print()
    decompress(save_path)


def download_data(url, data_dir, override, bloom_filter: set = None):
    bloom_filter = set() if bloom_filter is None else bloom_filter

    if url in bloom_filter:  # pragma: no cover
        return

    if url.endswith("/"):  # 以/结尾是文件夹，其余是文件
        _data_dir = path_append(data_dir, url.split('/')[-2], to_str=True)

        r = requests.get(url, timeout=30)
        r.raise_for_status()
        r.encoding = r.apparent_encoding
        soup = BeautifulSoup(r.text, "lxml")
        al = soup.find_all('a')
        for a in al:
            # 获得链接名
            h = a.get('href')
            if h[0] != '.':
                url_h = url + h
                if url_h not in bloom_filter:
                    download_data(url_h, _data_dir, override, bloom_filter)
        bloom_filter.add(url)

    else:
        os.makedirs(data_dir, exist_ok=True)
        save_path = path_append(data_dir, url.split('/')[-1], to_str=True)
        download_file(url, save_path, override)
        bloom_filter.add(url)

    return data_dir


def get_data(dataset, data_dir=DEFAULT_DATADIR, override=False, url_dict: dict = None):
    """
    Parameters
    ----------
    dataset: str
        数据集名
    data_dir: str
        数据存储目录
    override: bool
        是否覆盖已存在的文件
    url_dict:
        链接名称与链接映射

    Returns
    -------

    """
    url_dict = URL_DICT if not url_dict else url_dict
    if dataset in url_dict:
        url = url_dict[dataset]
    elif re.match("http(s?)://.*", dataset):
        url = dataset
    else:
        raise ValueError("%s is neither a valid dataset name nor an url" % dataset)

    try:
        return download_data(url, data_dir, override)
    except FileExistsError:
        return path_append(data_dir, url.split('/')[-1], to_str=True)


def list_resources():
    print("\n".join(URL_DICT))



# coding: utf-8
# create by tongshiwei on 2019-8-14

import io
import json

from tqdm import tqdm

__all__ = ["tl2json", "json2tl"]


def tl2json(src: str, tar: str, to_int=True, left_shift=False):
    """
    convert the dataset in `tl` sequence into `json` sequence

    .tl format
    The first line is the number of exercises a student attempted.
    The second line is the exercise tag sequence.
    The third line is the response sequence. ::

        15
        1,1,1,1,7,7,9,10,10,10,10,11,11,45,54
        0,1,1,1,1,1,0,0,1,1,1,1,1,0,0

    .json format
    Each sample contains several response elements, and each element is a two-element list.
    The first is the exercise tag and the second is the response. ::

        [[1,0],[1,1],[1,1],[1,1],[7,1],[7,1],[9,0],[10,0],[10,1],[10,1],[10,1],[11,1],[11,1],[45,0],[54,0]]

    """
    with open(src) as f, io.open(tar, "w", encoding="utf-8") as wf:
        for _ in tqdm(f):
            exercise_tags = f.readline().strip().strip(",").split(",")
            response_sequence = f.readline().strip().strip(",").split(",")
            if to_int:
                if not left_shift:
                    exercise_tags = list(map(int, exercise_tags))
                else:
                    exercise_tags = list(map(lambda x: int(x) - 1, exercise_tags))
                response_sequence = list(map(int, response_sequence))
            responses = list(zip(exercise_tags, response_sequence))
            print(json.dumps(responses), file=wf)


def json2tl(src, tar):
    with open(src) as f, io.open(tar, "w", encoding="utf-8") as wf:
        for line in tqdm(f):
            responses = json.loads(line)
            exercise_tags, response_sequence = zip(*responses)
            print(len(exercise_tags), file=wf)
            print(",".join(list(map(str, exercise_tags))), file=wf)
            print(",".join(list(map(str, response_sequence))), file=wf)


if __name__ == '__main__':
    json2tl("../data/junyi/student_log_kt.json.small.train", "../data/junyi/student_log_kt.json.small.train.tl")
    json2tl("../data/junyi/student_log_kt.json.small.test", "../data/junyi/student_log_kt.json.small.test.tl")



# coding: utf-8
# create by tongshiwei on 2019-7-5

from .junyi import build_knowledge_graph
from .KnowledgeTracing import select_n_most_frequent_students
from .main import extract_relations, build_json_sequence



# coding: utf-8
# 2019/12/6 @ tongshiwei

from EduData.DataSet.synthetic import transfer_synthetic_dataset
from longling import path_append


def test_synthetic_kt(shared_data_dir):
    src_dir = path_append(shared_data_dir, "synthetic", to_str=True)
    tar_dir = src_dir
    transfer_synthetic_dataset(src_dir, tar_dir)



# coding: utf-8
# create by tongshiwei on 2019/7/2

import pytest
from EduData import list_resources
from .conftest import get_data, test_url_dict


def test_download(shared_data_dir):
    get_data("tests", shared_data_dir, url_dict=test_url_dict)
    with pytest.raises(ValueError):
        get_data("*&53$#@")


def test_override(shared_data_dir):
    get_data(
        "http://base.ustc.edu.cn/data/tests/synthetic/naive_c5_q50_s4000_v0.csv",
        shared_data_dir,
    )
    get_data(
        "http://base.ustc.edu.cn/data/tests/synthetic/naive_c5_q50_s4000_v0.csv",
        shared_data_dir,
        override=True
    )
    get_data(
        "http://base.ustc.edu.cn/data/tests/synthetic/naive_c5_q50_s4000_v0.csv",
        shared_data_dir,
        override=False
    )


def test_list_resources():
    list_resources()



# coding: utf-8
# 2019/12/13 @ tongshiwei

"""
Using junyi test data as demo testing data
"""

import json

from longling import path_append, as_out_io

from EduData.Task.KnowledgeTracing.graph import (dense_graph,
                                                 correct_transition_count_graph, correct_transition_graph,
                                                 posterior_correct_probability_graph,
                                                 transition_graph, similarity_graph)


def test_graph(shared_data_dir, tmpdir):
    demo_response = [
        [[0, 1], [1, 0], [1, 1], [2, 0]],
        [[0, 0], [0, 0], [0, 1], [2, 0]],
        [[1, 1], [2, 0], [2, 1], [3, 1]],
        [[0, 1], [1, 1], [2, 0], [2, 1]],
        [[2, 0], [1, 0], [0, 1], [1, 1]],
    ]

    tmpfile = path_append(tmpdir, "demo.json", to_str=True)
    with as_out_io(tmpfile) as wf:
        for seq in demo_response:
            print(json.dumps(seq), file=wf)

    dense_graph_path = path_append(tmpdir, "dense_graph.json", to_str=True)
    _dense_graph = dense_graph(4, dense_graph_path)
    assert len(_dense_graph) == 12

    trans_graph = path_append(tmpdir, "transition_graph", to_str=True)
    transition_graph(4, tmpfile, tar=trans_graph)

    ctrans_count_graph = path_append(tmpdir, "correct_transition_count_graph", to_str=True)
    correct_transition_count_graph(4, tmpfile, tar=ctrans_count_graph)

    ctrans_graph = path_append(tmpdir, "correct_transition_graph", to_str=True)
    correct_transition_graph(4, tmpfile, tar=ctrans_graph)

    pcp_graph = path_append(tmpdir, "posterior_correct_probability_graph", to_str=True)
    posterior_correct_probability_graph(4, tmpfile, tar=pcp_graph)

    ctrans_sim = path_append(shared_data_dir, "correct_transition_sim_graph", to_str=True)
    similarity_graph(4, ctrans_graph, ctrans_sim)



# coding: utf-8
# 2019/12/6 @ tongshiwei

import os
import json
from tqdm import tqdm
from pathlib import PurePath
from longling import wf_open, path_append


def synthetic2json(src, tar):
    with open(src) as f, wf_open(tar) as wf:
        for line in tqdm(f, desc="%s -> %s" % (src, tar)):
            line = line.strip()
            if not line:  # pragma: no cover
                continue
            elems = line.split(",")
            print(json.dumps([[i, int(ans)] for i, ans in enumerate(elems)]), file=wf)


def transfer_synthetic_dataset(src_dir, tar_dir):
    for root, dirs, files in os.walk(src_dir):
        for filename in files:
            src = PurePath(path_append(root, filename))
            if src.suffix != ".csv":  # pragma: no cover
                continue
            tar = path_append(tar_dir, src.with_suffix(".json").name)
            synthetic2json(src, tar)


if __name__ == '__main__':
    transfer_synthetic_dataset("../../../data/synthetic", "./")



# coding: utf-8
# 2019/12/12 @ tongshiwei

import fileinput
import json
import warnings

import numpy as np
from longling import wf_open
from scipy.spatial.distance import cdist
from scipy.special import softmax
from tqdm import tqdm

__all__ = [
    "dense_graph",
    "correct_transition_count_graph",
    "correct_transition_graph", "transition_graph",
    "posterior_correct_probability_graph", "posterior_correct_transition_graph",
    "similarity_graph",
    "concurrence_graph", "correct_co_influence_graph"
]


def dense_graph(ku_num: int, tar=None, undirected: bool = False):
    """
    Dense graph where any two vertex have a link

    No self loop is reserved.

    Parameters
    ----------
    ku_num: int
    tar
    undirected

    Examples
    --------
    Target file is a json file, json.load can be used to read it.

    Demo of target file with undirected tag is False:
    [
        [0, 1],
        [0, 2],
        [1, 0],
        ...
        [2, 0],
        [2, 1]
    ]

    Demo of target file with undirected tag is True:
    [
        [0, 1],
        [1, 2],
        [0, 2]
    ]

    >>> dense_graph(3)
    [[0, 1], [0, 2], [1, 0], [1, 2], [2, 0], [2, 1]]
    >>> dense_graph(3, undirected=True)
    [[0, 1], [0, 2], [1, 2]]
    """
    _graph = []

    if undirected:
        for i in range(ku_num):
            for j in range(i + 1, ku_num):
                _graph.append([i, j])
    else:
        for i in range(ku_num):
            for j in range(ku_num):
                if i != j:
                    _graph.append([i, j])

    if tar is not None:
        with wf_open(tar) as wf:
            json.dump(_graph, wf, indent=2)
    return _graph


def _nan_divide(dividend_graph, divisor_graph, fill_na_with: (float, None) = 0., to_list=True):
    """

    Parameters
    ----------
    dividend_graph
    divisor_graph
    fill_na_with: float or None
    to_list

    Examples
    -------
    >>> a = [[10, 20], [30, 50]]
    >>> b = [[100, 200], [50, 100]]
    >>> _nan_divide(a, b)
    [[0.1, 0.1], [0.6, 0.5]]
    >>> a = [[0, 20], [30, 50]]
    >>> b = [[0, 200], [50, 100]]
    >>> _nan_divide(a, b)
    [[0.0, 0.1], [0.6, 0.5]]
    >>> _nan_divide(a, b, to_list=False)
    array([[0. , 0.1],
           [0.6, 0.5]])
    """
    _dividend_graph = np.asarray(dividend_graph)
    _divisor_graph = np.asarray(divisor_graph)

    _quotient_graph = _dividend_graph / _divisor_graph

    if fill_na_with is not None:
        _quotient_graph[np.isnan(_quotient_graph)] = fill_na_with

    if to_list:
        _quotient_graph = _quotient_graph.tolist()
    return _quotient_graph


def _row_normalize(count_graph, diagonal_value=None, normalized_by_softmax=False, skip_zero_row=False):
    """
    Normalize the count graph

    Parameters
    ----------
    count_graph
    diagonal_value: float or None
    normalized_by_softmax
    skip_zero_row

    Examples
    --------
    >>> _count_graph = [[0, 0, 0], [2, 8, 10], [82, 8, 0]]
    >>> _row_normalize(_count_graph, diagonal_value=0, skip_zero_row=True)
    [[0.0, 0.0, 0.0], [0.16666666666666666, 0.0, 0.8333333333333334], [0.9111111111111111, 0.08888888888888889, 0.0]]
    >>> _count_graph = [[0, 0], [2, 8]]
    >>> _row_normalize(_count_graph)
    [[0.5, 0.5], [0.2, 0.8]]
    >>> _count_graph = [[2, 3], [2, 8]]
    >>> _row_normalize(_count_graph)
    [[0.4, 0.6], [0.2, 0.8]]
    >>> _count_graph = [[0, 0], [2, 8]]
    >>> _row_normalize(_count_graph, normalized_by_softmax=True)
    [[0.5, 0.5], [0.002472623156634775, 0.9975273768433656]]
    >>> _count_graph = [[2, 3], [2, 8]]
    >>> _row_normalize(_count_graph, normalized_by_softmax=True)
    [[0.26894142136999505, 0.7310585786300048], [0.002472623156634775, 0.9975273768433656]]
    >>> _count_graph = [[0, 0, 0], [2, 8, 10], [82, 8, 0]]
    >>> _row_normalize(_count_graph, diagonal_value=0)
    [[0.0, 0.5, 0.5], [0.16666666666666666, 0.0, 0.8333333333333334], [0.9111111111111111, 0.08888888888888889, 0.0]]
    >>> _count_graph = [[0, 0, 0], [2, 8, 10], [82, 8, 0]]
    >>> import numpy as np
    >>> np.asarray(_row_normalize(_count_graph, normalized_by_softmax=True))
    array([[3.33333333e-01, 3.33333333e-01, 3.33333333e-01],
           [2.95387223e-04, 1.19167711e-01, 8.80536902e-01],
           [1.00000000e+00, 7.28129018e-33, 2.44260074e-36]])
    >>> np.asarray(_row_normalize(_count_graph, normalized_by_softmax=True, diagonal_value=0.0))
    array([[0.00000000e+00, 5.00000000e-01, 5.00000000e-01],
           [3.35350130e-04, 0.00000000e+00, 9.99664650e-01],
           [1.00000000e+00, 7.28129018e-33, 0.00000000e+00]])
    """
    _graph = np.asarray(count_graph)
    zero_rows_indices = None

    if skip_zero_row:
        zero_rows_indices = np.sum(_graph, axis=-1) == 0

    if normalized_by_softmax:
        if diagonal_value is None:
            _graph = softmax(_graph, axis=-1)
        else:
            _graph = _graph.astype(float)
            np.fill_diagonal(_graph, -np.inf)
            _graph = softmax(_graph, axis=-1)
            if diagonal_value is not None:
                np.fill_diagonal(_graph, diagonal_value)
    else:
        _offset_graph = _graph + 1e-50
        if diagonal_value is not None:
            np.fill_diagonal(_offset_graph, 0.0)
            _graph = (_offset_graph.T / _offset_graph.sum(axis=-1)).T
            _graph[_graph <= 1e-40] = 0.0
            np.fill_diagonal(_graph, diagonal_value)
        else:
            _graph = (_offset_graph.T / _offset_graph.sum(axis=-1)).T
            _graph[_graph <= 1e-40] = 0.0

    if skip_zero_row:
        _graph[zero_rows_indices] = 0.0

    return _graph.tolist()


def _output_graph(graph, tar):
    ku_num = len(graph)

    _graph = []

    for i in range(ku_num):
        for j in range(ku_num):
            if i != j and not np.isnan(graph[i][j]) and graph[i][j] > 0:
                _graph.append([i, j, graph[i][j]])

    with wf_open(tar) as wf:
        json.dump(_graph, wf, indent=2)


def _correct_transition_count_graph(count_graph, seq):
    pre_c = None
    for eid, r in seq:
        if pre_c is not None:
            if eid != pre_c and r == 1:
                count_graph[pre_c][eid] += 1
            elif r == 1:
                pass
        if r == 1:
            pre_c = eid
        else:
            pre_c = None


def correct_transition_count_graph(ku_num, *src, tar=None, input_is_file=True):
    """

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file

    Examples
    -------
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> correct_transition_count_graph(3, _seq, input_is_file=False)
    [[0, 0, 1], [0, 0, 1], [0, 0, 0]]
    >>> _seq = [[[0, 1], [1, 1], [1, 1], [2, 1]]]
    >>> correct_transition_count_graph(3, _seq, input_is_file=False)
    [[0, 1, 0], [0, 0, 1], [0, 0, 0]]
    """
    count_graph = [[0] * ku_num for _ in range(ku_num)]

    if input_is_file:
        with fileinput.input(files=src) as f:
            for line in tqdm(f, "constructing coorect transition graph"):
                if not line.strip():  # pragma: no cover
                    continue
                seq = json.loads(line)
                _correct_transition_count_graph(count_graph, seq)
    else:
        for seqs in src:
            for seq in seqs:
                _correct_transition_count_graph(count_graph, seq)

    if tar is not None:
        _output_graph(count_graph, tar)

    return count_graph


def correct_transition_graph(ku_num, *src, tar=None, input_is_file=True, diagonal_value=0.):
    """
    When a concept is mastered, how much probability is it to be transferred to another concept.

    For example,

    ```
    [[0, 1], [1, 0], [1, 1], [2, 1]]
    [[2, 0], [1, 0], [0, 1], [2, 1]]
    ```
    When concept #0 is mastered (i.e., 1st in seq #1, 3rd in seq #2),
    only concept # 2 can be mastered (4th in seq #2).
    Thus, the transition probabilty for concept #0 is [0, 0, 1],
    which mastering concept #0 can influence mastering concept #2 more thant concept #1.

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file
    diagonal_value

    Returns
    -------

    Examples
    -------
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> correct_transition_graph(3, _seq, input_is_file=False)
    [[0.0, 0.0, 1.0], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    >>> _seq = [[[0, 1], [1, 1], [1, 1], [2, 1]]]
    >>> correct_transition_graph(3, _seq, input_is_file=False)
    [[0.0, 1.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    """
    count_graph = correct_transition_count_graph(ku_num, *src, tar=None, input_is_file=input_is_file)
    _transition_graph = _row_normalize(count_graph, diagonal_value, skip_zero_row=True)

    if tar is not None:
        _output_graph(_transition_graph, tar)

    return _transition_graph


def transition_graph(ku_num, *src, tar=None, input_is_file=True, diagonal_value=0.):
    """
    When a concept is learned, how much probability does another concept appear.

    For example,

    ```
    [[0, 1], [1, 0], [1, 1], [2, 1]]
    [[2, 0], [1, 0], [0, 1], [2, 1]]
    ```
    When concept #0 is learned (i.e., 1st in seq #1, 3rd in seq #2),
    concept #2 and #1 could appear (2nd in seq #1, 4th in seq #2)
    Thus, the transition probabilty for concept #0 is [0, 0.5, 0.5].

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file
    diagonal_value

    Returns
    -------

    Examples
    --------
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> transition_graph(3, _seq, input_is_file=False)
    [[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.0, 1.0, 0.0]]
    >>> _seq = [[[0, 1], [1, 1], [1, 1], [2, 1]]]
    >>> transition_graph(3, _seq, input_is_file=False)
    [[0.0, 1.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    """
    count_graph = [[0] * ku_num for _ in range(ku_num)]

    def __transition_graph(_seq):
        pre = None
        for eid, _ in _seq:
            if pre is not None:
                if eid != pre:
                    count_graph[pre][eid] += 1
                else:
                    # count_graph[pre][eid] += 1
                    pass
            pre = eid

    if input_is_file:
        with fileinput.input(files=src) as f:
            for line in tqdm(f, "constructing transition graph"):
                if not line.strip():  # pragma: no cover
                    continue
                seq = json.loads(line)
                __transition_graph(seq)

    else:
        for seqs in src:
            for seq in seqs:
                __transition_graph(seq)

    _transition_graph = _row_normalize(count_graph, diagonal_value, skip_zero_row=True)
    if tar is not None:
        _output_graph(_transition_graph, tar)

    return _transition_graph


def posterior_correct_probability_graph(ku_num, *src, tar=None, input_is_file=True, fill_na_with=0.):
    """
    When a concept is mastered, how much probability is another concept correctly answered.

    For example,

    ```
    [[0, 1], [1, 1], [1, 1], [2, 1]]
    [[2, 0], [1, 0], [0, 1], [2, 1]]
    ```
    When concept #0 is mastered (i.e., 1st in seq #1, 3rd in seq #2),
    concept #1 and # 2 can both be mastered (1th in seq # 1, 4th in seq #2).
    Thus, the posterior correct probability for concept #0 is [0, 1, 1].

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file
    fill_na_with

    Returns
    -------
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> posterior_correct_probability_graph(3, _seq, input_is_file=False)
    [[0.0, 1.0, 1.0], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    """
    count_graph = [[0] * ku_num for _ in range(ku_num)]
    correct_count_graph = [[0] * ku_num for _ in range(ku_num)]

    def __posterior_correct_graph(seq):
        pre_c = None
        for eid, r in seq:
            if pre_c is not None:
                if eid != pre_c:
                    count_graph[pre_c][eid] += 1
                    if r == 1:
                        correct_count_graph[pre_c][eid] += 1
                    elif r == 0:
                        correct_count_graph[pre_c][eid] += 1
            if r == 1:
                pre_c = eid
            else:
                pre_c = None

    assert src

    if input_is_file:
        with fileinput.input(files=src) as f:
            for line in tqdm(f, "constructing correct transition graph"):
                if not line.strip():  # pragma: no cover
                    continue
                seq = json.loads(line)
                __posterior_correct_graph(seq)
    else:
        for seqs in src:
            for seq in seqs:
                __posterior_correct_graph(seq)

    _posterior_correct_graph = _nan_divide(correct_count_graph, count_graph, fill_na_with)
    if tar is not None:
        _output_graph(_posterior_correct_graph, tar)

    return _posterior_correct_graph


def posterior_correct_transition_graph(ku_num, *src, tar=None, input_is_file=True, diagonal_value=None):
    """
    Correct transition graph based on posterior correct graph

    For example,

    ```
    [[0, 1], [1, 1], [1, 1], [2, 1]]
    [[2, 0], [1, 0], [0, 1], [2, 1]]
    ```
    When concept #0 is mastered (i.e., 1st in seq #1, 3rd in seq #2),
    concept #1 and # 2 can both be mastered (1th in seq # 1, 4th in seq #2).
    Thus, the posterior correct probability for concept #0 is [0, 1, 1].

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file
    diagonal_value

    Returns
    -------
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> posterior_correct_transition_graph(3, _seq, input_is_file=False)
    [[0.0, 0.5, 0.5], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    >>> _seq = [[[0, 1], [1, 0], [1, 1], [2, 1]], [[2, 0], [1, 0], [0, 1], [2, 1]]]
    >>> posterior_correct_transition_graph(3, _seq, input_is_file=False)
    [[0.0, 0.5, 0.5], [0.0, 0.0, 1.0], [0.0, 0.0, 0.0]]
    """
    return _row_normalize(
        posterior_correct_probability_graph(
            ku_num,
            *src,
            tar=tar,
            input_is_file=input_is_file
        ),
        diagonal_value=diagonal_value,
        skip_zero_row=True
    )


def correct_co_influence_graph(ku_num, *src, tar=None, input_is_file=True):  # pragma: no cover
    """

    Co-influence graph

    A co-influence pair is defined as two vertexes
    that the sum of transition count is large and the difference is small.

    Diagonal_value is always 0

    Parameters
    ----------
    ku_num
    src
    tar
    input_is_file

    Examples
    --------
    >>> _seq = [
    ...     [[0, 1], [1, 0], [1, 1], [2, 0]],
    ...     [[0, 1], [1, 1], [2, 0], [2, 1]],
    ...     [[2, 1], [2, 1], [1, 1], [2, 0]],
    ...     [[1, 0], [0, 1], [0, 1], [2, 0]],
    ...     [[2, 0], [1, 1], [0, 1], [2, 1]],
    ... ]
    >>> correct_co_influence_graph(3, _seq, input_is_file=False)
    array([[0., 1., 0.],
           [1., 0., 0.],
           [0., 0., 0.]])
    """
    warnings.warn("do not use this function due to the lack of support from theory")
    count_graph = correct_transition_count_graph(ku_num, *src, tar=None, input_is_file=input_is_file)

    for i in range(ku_num):
        for j in range(i + 1, ku_num):
            count_graph[i][j] = count_graph[i][j] + count_graph[j][i] / (
                    abs(count_graph[i][j] - count_graph[j][i]) + 1e-8)
            count_graph[j][i] = count_graph[i][j]

    count_graph = np.asarray(count_graph)
    _concurrence_graph = softmax(count_graph) * 2

    if tar is not None:
        _output_graph(_concurrence_graph, tar)

    return _concurrence_graph


def concurrence_graph(ku_num, *src, tar):  # pragma: no cover
    warnings.warn("do not use this function due to the lack of support from theory")

    count_graph = [[0] * ku_num for _ in range(ku_num)]

    with fileinput.input(files=src) as f:
        for line in tqdm(f, "constructing concurrence graph"):
            if not line.strip():  # pragma: no cover
                continue
            seq = json.loads(line)
            pre = None
            for eid, _ in seq:
                if pre is not None:
                    if eid != pre:
                        try:
                            count_graph[pre][eid] += 1
                            count_graph[eid][pre] += 1
                        except IndexError:
                            print(pre, eid)
                            exit(-1)
                    else:
                        # count_graph[pre][eid] += 1
                        pass
                pre = eid

    count_graph = np.asarray(count_graph)
    _concurrence_graph = softmax(count_graph) * 2
    _output_graph(_concurrence_graph, tar)


def similarity_graph(ku_num, src_graph, tar):
    """construct similarity graph based on transition graph"""
    with open(src_graph) as f:
        _transitions = json.load(f)

    _transition_graph = [[0] * ku_num for _ in range(ku_num)]

    for id1, id2, value in _transitions:
        _transition_graph[id1][id2] = float(value)

    _transition_graph = np.asarray(_transition_graph)
    _dist_graph = 1 - cdist(_transition_graph, _transition_graph, 'cosine')

    _output_graph(_dist_graph, tar)



# coding: utf-8
# 2019/8/23 @ tongshiwei

import fire

from EduData.DataSet.download_data.download_data import get_data, list_resources
from EduData.Task.KnowledgeTracing.format import tl2json, json2tl
from EduData.Task.KnowledgeTracing.statistics import analysis_records, analysis_edges
from longling.ML.toolkit.dataset import train_valid_test, kfold
from EduData.DataSet.junyi import extract_relations, build_json_sequence
from EduData.DataSet.EdNet import build_interactions, select_n_most_active
from EduData.Task.KnowledgeTracing.graph import dense_graph, transition_graph, correct_transition_graph
from EduData.Task.KnowledgeTracing.graph import similarity_graph, concurrence_graph, correct_co_influence_graph


def cli():  # pragma: no cover
    fire.Fire(
        {
            "download": get_data,
            "ls": list_resources,
            "tl2json": tl2json,
            "json2tl": json2tl,
            "kt_stat": analysis_records,
            "edge_stat": analysis_edges,
            "train_valid_test": train_valid_test,
            "kfold": kfold,
            "dataset": {
                "junyi": {
                    "kt": {
                        "extract_relations": extract_relations,
                        "build_json_sequence": build_json_sequence,
                    }
                },
                "ednet": {
                    "kt": {
                        "build_json_sequence": build_interactions,
                        "select_n": select_n_most_active,
                    }
                }
            },
            "graph": {
                "ccon": correct_co_influence_graph,
                "con": concurrence_graph,
                "dense": dense_graph,
                "trans": transition_graph,
                "ctrans": correct_transition_graph,
                "sim": similarity_graph,
            }
        }
    )



# coding: utf-8
# 2019/12/17 @ tongshiwei

import csv


def get_question_id(question_str):
    """
    Examples
    --------
    >>> get_question_id("q123")
    123
    """
    return int(question_str.lstrip("q"))


class Judgement(object):
    def __init__(self, questions_csv):
        self.question = [None] * (18143 + 1)
        with open(questions_csv) as f:
            f.readline()
            for i, line in enumerate(csv.reader(f, delimiter=",")):
                _id = self.get_question_id(line[0])
                correct_answer = line[3]
                self.question[_id] = (i, correct_answer)

    @staticmethod
    def get_question_id(question_id):
        if isinstance(question_id, str):
            question_id = get_question_id(question_id)
        return question_id

    def is_correct(self, question_id: (str, int), user_answer: str):
        question_id = self.get_question_id(question_id)

        if self.question[question_id] is None:
            raise ValueError("Unknown question")

        return True if self.question[question_id][1] == user_answer else False

    def __call__(self, question_id: str, user_answer: str):
        question_id = self.get_question_id(question_id)

        if self.question[question_id] is None:
            raise ValueError("Unknown question")

        _id, ground_truth = self.question[question_id]
        return _id, 1 if ground_truth == user_answer else 0



# coding: utf-8
# create by tongshiwei on 2019-7-5

"""
This script is used to convert the original junyi dataset into json sequence, which can be applied in kt task.
"""

__all__ = ["select_n_most_frequent_students"]

import csv
import json

from longling import wf_open
from longling.lib.candylib import as_list
from tqdm import tqdm


def _read(source: str, ku_dict: str) -> dict:
    """
    Read the learners' interaction records and classify them by user id and session id.
    In the same time, the exercise name will be converted to id.

    Notes
    -----
    Require big memory to run this function.
    """

    outcome = {
        "INCORRECT": 0,
        "CORRECT": 1,
        "HINT": 0,
    }

    students = {}

    with open(ku_dict) as f:
        ku_dict = json.load(f)

    with open(source) as f:
        f.readline()
        for line in tqdm(csv.reader(f, delimiter='\t'), "reading data"):
            student, session, exercise = line[0], line[1], ku_dict[line[-5]],
            correct, timestamp = outcome[line[10]], line[8]
            if student not in students:
                students[student] = {}
            if session not in students[student]:
                students[student][session] = []

            students[student][session].append([int(timestamp), exercise, correct])
    return students


def _write(students, target):
    with wf_open(target) as wf:
        for student_id, sessions in tqdm(students.items(), "writing -> %s" % target):
            for session_id, exercises in sessions.items():
                exercises.sort(key=lambda x: x[0])
                exercise_response = [(exercise[1], exercise[2]) for exercise in exercises]
                print(json.dumps(exercise_response), file=wf)


def _frequency(students):
    frequency = {}
    for student_id, sessions in tqdm(students.items(), "calculating frequency"):
        frequency[student_id] = sum([len(session) for session in sessions])
    return sorted(frequency.items(), key=lambda x: x[1], reverse=True)


def get_n_most_frequent_students(students, n=None, frequency: list = None):
    frequency = _frequency(students) if frequency is None else frequency
    __frequency = frequency if n is None else frequency[:n]
    _students = {}
    for _id, _ in __frequency:
        _students[_id] = students[_id]
    return _students


def select_n_most_frequent_students(source: str, target_prefix: str, ku_dict_path: str, n: (int, list)):
    """None in n means select all students"""
    n_list = as_list(n)
    students = _read(source, ku_dict_path)
    frequency = _frequency(students)
    for _n in n_list:
        _write(get_n_most_frequent_students(students, _n, frequency), target_prefix + "%s" % _n)



# coding: utf-8
# 2019/12/17 @ tongshiwei

import json
import pytest
from longling import path_append
from EduData.DataSet.EdNet.utils import Judgement
from EduData.DataSet.EdNet.KnowledgeTracing import build_interactions, csv2interactions, select_n_most_active


def test_utils(shared_data_dir):
    question_csv_path = path_append(shared_data_dir, "tests", "EdNet", "contents", "questions.csv")

    judgement = Judgement(question_csv_path)

    assert judgement.is_correct("q1", "b") is True
    assert judgement.is_correct("1", "c") is False

    with pytest.raises(ValueError):
        judgement.is_correct("q12000", "d")

    with pytest.raises(ValueError):
        judgement("q12000", "a")

    assert judgement("q1", "b") == (0, 1)


def test_build_interactions(shared_data_dir):
    question_csv_path = path_append(shared_data_dir, "tests", "EdNet", "contents", "questions.csv")
    users_dir = path_append(shared_data_dir, "tests", "EdNet", "KT1")

    user_csv = path_append(users_dir, "u1.csv")

    judgement = Judgement(question_csv_path)

    interactions = csv2interactions(user_csv, judgement)

    assert interactions[0] == [5011, 0]

    tar = path_append(shared_data_dir, "tests", "EdNet", "KT1", "data", "kt.json")
    build_interactions(users_dir, question_csv_path, tar)

    with open(tar) as f:
        assert json.loads(f.readline())[0] == [5011, 0]
        assert len(json.loads(f.readline())) == 16

    tar2 = path_append(tar, to_str=True) + ".%s" % 1
    select_n_most_active(tar, tar2, 1)

    with open(tar2) as f:
        assert json.loads(f.readline())[0] == [5011, 0]



# coding: utf-8
# create by tongshiwei on 2019-8-14

from longling import as_out_io

from EduData.Task.KnowledgeTracing import tl2json, json2tl

TL_STR = """15
1,1,1,1,7,7,9,10,10,10,10,11,11,45,54
0,1,1,1,1,1,0,0,1,1,1,1,1,0,0
15
1,1,1,1,7,7,9,10,10,10,10,11,11,45,54
0,1,1,1,1,1,0,0,1,1,1,1,1,0,0
15
1,1,1,1,7,7,9,10,10,10,10,11,11,45,54
0,1,1,1,1,1,0,0,1,1,1,1,1,0,0
"""


def test_tl_json(tmpdir):
    tl_file = str(tmpdir / "demo.tl")
    json_file = str(tmpdir / "demo.json")

    with as_out_io(tl_file) as wf:
        print(TL_STR, file=wf, end='')

    tl2json(tl_file, json_file)
    json2tl(json_file, tl_file)

    with open(tl_file) as f:
        assert f.read() == TL_STR

    tl2json(tl_file, json_file, left_shift=True)



# coding: utf-8
# 2019/8/24 @ tongshiwei

__all__ = ["analysis_records"]

import networkx as nx
import numpy as np
import pandas as pd
import fileinput
from tqdm import tqdm
import json


def _report(total, correct):
    keep_ids, total_first_answer_num = list(zip(*[[i, n] for i, n in enumerate(total) if n > 0]))
    total_correctly_first_answer_num = [n for i, n in enumerate(correct) if i in keep_ids]
    total_first_answer_num = np.asarray(total_first_answer_num)
    total_correctly_first_answer_num = np.asarray(total_correctly_first_answer_num)
    correctly_answer_portion = total_correctly_first_answer_num / total_first_answer_num
    print(total_first_answer_num.sum())
    print(total_correctly_first_answer_num.sum())
    print(total_correctly_first_answer_num.sum() / total_first_answer_num.sum())
    print(correctly_answer_portion.max(), correctly_answer_portion.min(),
          correctly_answer_portion.sum() / len(correctly_answer_portion))


def correctly_answer(ku_num, *filenames, prerequsite_graph=None, similarity_graph=None):
    total_first_answer_num = [0] * ku_num
    total_correctly_first_answer_num = [0] * ku_num

    pre_total_first_answer_num = [0] * ku_num
    pre_total_correctly_first_answer_num = [0] * ku_num

    sim_total_first_answer_num = [0] * ku_num
    sim_total_correctly_first_answer_num = [0] * ku_num

    graph_total_first_answer_num = [0] * ku_num
    graph_total_correctly_first_answer_num = [0] * ku_num

    with open(prerequsite_graph) as f:
        pgraph = nx.DiGraph()
        pgraph.add_nodes_from([i for i in range(ku_num)])
        pgraph.add_edges_from(json.load(f))
        pre_path = nx.shortest_path(pgraph)

    with open(similarity_graph) as f:
        sgraph = nx.Graph()
        sgraph.add_nodes_from([i for i in range(ku_num)])
        sgraph.add_edges_from([edge[:2] for edge in json.load(f) if edge[2] >= 5.0])
        sim_path = nx.shortest_path(sgraph)

    with fileinput.input(files=filenames) as f:
        for line in tqdm(f):
            records = json.loads(line)
            pre_set = set()
            pc_set = set()
            pre_ku = None
            for _id, _c in records:
                if pre_ku is None:
                    total_first_answer_num[_id] += 1
                    if _c == 1:
                        total_correctly_first_answer_num[_id] += 1
                elif _id != pre_ku:
                    total_first_answer_num[_id] += 1
                    if _c == 1:
                        total_correctly_first_answer_num[_id] += 1

                    in_pre = False
                    in_sim = False
                    for pre in pc_set:
                        if in_pre and in_sim:
                            break
                        if _id in pre_path[pre] and len(pre_path[pre][_id]) <= 2:
                            in_pre = True
                        if _id in sim_path[pre] and len(sim_path[pre][_id]) <= 2:
                            in_sim = True
                    if in_pre:
                        pre_total_first_answer_num[_id] += 1
                        if _c == 1:
                            pre_total_correctly_first_answer_num[_id] += 1
                    if in_sim:
                        sim_total_first_answer_num[_id] += 1
                        if _c == 1:
                            sim_total_correctly_first_answer_num[_id] += 1
                    if in_pre or in_sim:
                        graph_total_first_answer_num[_id] += 1
                        if _c == 1:
                            graph_total_correctly_first_answer_num[_id] += 1

                    pre_set.add(_id)
                    if _c == 1:
                        pc_set.add(_id)
                pre_ku = _id
    print("total")
    _report(total_first_answer_num, total_correctly_first_answer_num)
    print("pre")
    _report(pre_total_first_answer_num, pre_total_correctly_first_answer_num)
    print("sim")
    _report(sim_total_first_answer_num, sim_total_correctly_first_answer_num)
    print("graph")
    _report(graph_total_first_answer_num, graph_total_correctly_first_answer_num)


def correctly_answer2(ku_num, *filenames, prerequsite_graph=None, similarity_graph=None):
    total_first_answer_num = [0] * ku_num
    total_correctly_first_answer_num = [0] * ku_num

    pre_total_first_answer_num = [0] * ku_num
    pre_total_correctly_first_answer_num = [0] * ku_num

    sim_total_first_answer_num = [0] * ku_num
    sim_total_correctly_first_answer_num = [0] * ku_num

    graph_total_first_answer_num = [0] * ku_num
    graph_total_correctly_first_answer_num = [0] * ku_num

    with open(prerequsite_graph) as f:
        pgraph = nx.DiGraph()
        pgraph.add_nodes_from([i for i in range(ku_num)])
        pgraph.add_edges_from(json.load(f))
        pre_path = nx.shortest_path(pgraph)

    with open(similarity_graph) as f:
        sgraph = nx.Graph()
        sgraph.add_nodes_from([i for i in range(ku_num)])
        sgraph.add_edges_from([edge[:2] for edge in json.load(f) if edge[2] >= 5.0])
        sim_path = nx.shortest_path(sgraph)

    with fileinput.input(files=filenames) as f:
        for line in tqdm(f):
            records = json.loads(line)
            pre = None
            pre_c = None
            for _id, _c in records:
                if pre is None:
                    total_first_answer_num[_id] += 1
                    if _c == 1:
                        total_correctly_first_answer_num[_id] += 1
                elif _id != pre:
                    total_first_answer_num[_id] += 1
                    if _c == 1:
                        total_correctly_first_answer_num[_id] += 1

                    if pre_c == 1:
                        in_graph = False
                        if _id in pre_path[pre] and len(pre_path[pre][_id]) <= 2:
                            in_graph = True
                            pre_total_first_answer_num[_id] += 1
                            if _c == 1:
                                pre_total_correctly_first_answer_num[_id] += 1
                        if _id in sim_path[pre] and len(sim_path[pre][_id]) <= 2:
                            in_graph = True
                            sim_total_first_answer_num[_id] += 1
                            if _c == 1:
                                sim_total_correctly_first_answer_num[_id] += 1
                        if in_graph:
                            graph_total_first_answer_num[_id] += 1
                            if _c == 1:
                                graph_total_correctly_first_answer_num[_id] += 1
                pre = _id
                pre_c = _c

    print("total")
    _report(total_first_answer_num, total_correctly_first_answer_num)
    print("pre")
    _report(pre_total_first_answer_num, pre_total_correctly_first_answer_num)
    print("sim")
    _report(sim_total_first_answer_num, sim_total_correctly_first_answer_num)
    print("graph")
    _report(graph_total_first_answer_num, graph_total_correctly_first_answer_num)


def analysis_records(*source):
    ku_set = set()
    records_num = 0
    seq_count = 0
    correct_num = 0
    with fileinput.input(files=source) as f:
        for line in tqdm(f, "doing statistics"):
            seq_count += 1
            responses = json.loads(line)
            records_num += len(responses)
            correct_num += len([r[1] for r in responses if int(r[1]) == 1])
            ku_set.update(set([_id for _id, _ in responses]))

    print("in %s" % list(source))
    print("knowledge units number: %s" % len(ku_set))
    print("min index: %s; max index: %s" % (min(ku_set), max(ku_set)))
    print("records number: %s" % records_num)
    print("correct records number: %s" % correct_num)
    print("the number of sequence: %s" % seq_count)


def analysis_edges(src, threshold=None):
    edge_num = 0
    average_weight = []

    with open(src) as f:
        graph_edges = json.load(f)

    for edge in graph_edges:
        if len(edge) == 2:
            edge_num += 1
        elif len(edge) >= 3:
            if threshold is None:
                edge_num += 1
                average_weight.append(edge[2])
            elif edge[2] >= threshold:
                edge_num += 1
        else:  # pragma: no cover
            raise ValueError("each edge in src should have at least two element")

    print("in %s" % src)
    print("%s edges" % edge_num)
    if average_weight:
        print(pd.Series(average_weight).describe())


if __name__ == '__main__':
    root = "../../../data/junyi/"
    correctly_answer(
        835,
        root + "train.json", root + "test.json",
        prerequsite_graph=root + "prerequisite.json",
        similarity_graph=root + "similarity.json"
    )



# coding: utf-8
# create by tongshiwei on 2019/6/25

from setuptools import setup, find_packages

test_deps = [
    'pytest>=4',
    'pytest-cov>=2.6.0',
    'pytest-flake8',
]

setup(
    name='EduData',
    version='0.0.10',
    extras_require={
        'test': test_deps,
    },
    packages=find_packages(),
    python_requires='>=3.6',
    long_description='Refer to full documentation https://github.com/bigdata-ustc/EduData/blob/master/README.md'
                     ' for detailed information.',
    description='This project aims to '
                'provide convenient interfaces for downloading and preprocessing dataset in education.',
    install_requires=[
        'tqdm',
        'networkx',
        'longling[ml]>=1.3.15',
        'requests',
        'bs4',
        'rarfile',
        'pandas',
        'fire',
        'lxml',
        'numpy',
        'scipy>=1.2.0',
        'scikit-learn>=0.23.2',
    ],  # And any other dependencies foo needs
    entry_points={
        "console_scripts": [
            "edudata = EduData.main:cli",
        ],
    },
    classifier=[
        "Programming Language :: Python :: 3",
    ]
)



# coding: utf-8
# create by tongshiwei on 2019/7/2

"""
This script is used to build the map dict (ku_name -> idx) extract some relations from the original junyi dataset.
"""
__all__ = ["build_knowledge_graph"]

import codecs
import csv
import json

import networkx as nx
import pandas
from longling import wf_open, config_logging, path_append
from tqdm import tqdm

logger = config_logging(logger="junyi", console_log_level="info")


def build_ku_dict(source, target):
    with codecs.open(source, encoding="utf-8") as f, wf_open(target) as wf:
        f.readline()
        idx = 0
        vertex_dict = {}
        for line in tqdm(csv.reader(f)):
            if line[0] not in vertex_dict:
                vertex_dict[line[0]] = idx
                idx += 1
        logger.info("vertex num: %s" % len(vertex_dict))
        json.dump(vertex_dict, wf, indent=2)


def extract_prerequisite(source, target, ku_dict):
    """in target: (A, B) means predecessor --> successor"""
    with codecs.open(source, encoding="utf-8") as f, open(ku_dict) as kf, wf_open(target) as wf:
        ku_dict = json.load(kf)

        prerequisite_edges = []
        f.readline()
        for line in tqdm(csv.reader(f)):
            if not line[2]:
                continue
            successor = ku_dict[line[0]]
            for prerequisite in line[2].split(','):
                predecessor = ku_dict[prerequisite]
                if predecessor == successor:
                    continue
                if predecessor == 61 and successor == 498:
                    # there is a loop 498 -> 510 -> 61 -> 498 in original data
                    continue
                prerequisite_edges.append((predecessor, successor))

        logger.info("prerequisite edges: %s" % len(prerequisite_edges))

        # clean the loop in prerequisite graph

        graph = nx.DiGraph()
        graph.add_edges_from(prerequisite_edges)
        assert not list(nx.algorithms.simple_cycles(graph)), "loop in DiGraph"

        json.dump(prerequisite_edges, wf, indent=2)


def merge_relationship_annotation(sources, target):
    with wf_open(target) as wf:
        with codecs.open(sources[0]) as f:
            for line in f:
                wf.write(line)
        with codecs.open(sources[1]) as f:
            f.readline()
            for line in f:
                wf.write(line)
    return target


def extract_similarity(source, target, ku_dict):
    """
    In target: (A, B, v) means A is similar with B in v degree.
    If v is small, A and B should be considered as not similar.
    """
    similarity = []
    with codecs.open(source, encoding="utf-8") as f, open(ku_dict) as kf, wf_open(target) as wf:
        f.readline()
        ku_dict = json.load(kf)
        for line in csv.reader(f):
            similarity.append((ku_dict[line[0]], ku_dict[line[1]], float(line[2])))

        logger.info("similarity edges: %s" % len(similarity))

        logger.info(pandas.Series([sim[-1] for sim in similarity]).describe())
        json.dump(similarity, wf, indent=2)


def extract_difficulty(source, target, ku_dict):
    """
    In target: (A, B, v) means A is similar with B in v degree.
    If v is small, A and B should be considered as not similar.
    """
    difficulty = []
    with codecs.open(source, encoding="utf-8") as f, open(ku_dict) as kf, wf_open(target) as wf:
        f.readline()
        ku_dict = json.load(kf)
        for line in csv.reader(f):
            difficulty.append((ku_dict[line[0]], ku_dict[line[1]], float(line[4])))

        logger.info("edges: %s" % len(difficulty))

        logger.info(pandas.Series([sim[-1] for sim in difficulty]).describe())
        json.dump(difficulty, wf, indent=2)


def build_knowledge_graph(src_root: str, tar_root: (str, None) = None,
                          ku_dict_path: str = None,
                          prerequisite_path: (str, None) = None,
                          similarity_path: (str, None) = None,
                          difficulty_path: (str, None) = None):
    tar_root = tar_root if tar_root is not None else src_root
    exercise_src = path_append(src_root, "junyi_Exercise_table.csv")

    assert ku_dict_path is not None

    relation_src = merge_relationship_annotation(
        [path_append(src_root, "relationship_annotation_{}.csv".format(name)) for name in ["testing", "training"]],
        path_append(src_root, "relationship_annotation.csv")
    )
    ku_dict_path = path_append(tar_root, ku_dict_path)
    build_ku_dict(exercise_src, ku_dict_path)

    if prerequisite_path is not None:
        prerequisite_path = path_append(tar_root, prerequisite_path)
        extract_prerequisite(exercise_src, prerequisite_path, ku_dict_path)

    if similarity_path is not None:
        similarity_path = path_append(tar_root, "similarity.json")
        extract_similarity(relation_src, similarity_path, ku_dict_path)

    if difficulty_path is not None:
        difficulty_path = path_append(tar_root, "difficulty.json")
        extract_difficulty(relation_src, difficulty_path, ku_dict_path)



# coding: utf-8
# 2019/12/17 @ tongshiwei

import heapq
import json
from longling import path_append, wf_open
import re
import os
import csv
from tqdm import tqdm
from .utils import Judgement

__all__ = ["csv2interactions", "build_interactions"]


def csv2interactions(src: str, judgement: Judgement):
    interactions = []
    with open(src) as f:
        f.readline()
        for line in csv.reader(f, delimiter=","):
            _question_id = line[2]
            _answer = line[3]
            interactions.append(list(judgement(_question_id, _answer)))
    return interactions


def build_interactions(users_dir, questions_csv, tar):
    judgement = Judgement(questions_csv)

    with wf_open(tar) as wf:
        for root, dirs, files in os.walk(users_dir):
            for filename in tqdm(files, "building interactions"):
                if re.match("u.*\.csv", filename):
                    interactions_seq = csv2interactions(path_append(root, filename, to_str=True), judgement)
                    print(json.dumps(interactions_seq), file=wf)


def select_n_most_active(src, tar, n):
    lengths = []
    with open(src) as f:
        for i, line in tqdm(enumerate(f), "evaluating length of each row"):
            lengths.append([i, len(json.loads(line))])

    selected_idx = set(list(zip(*heapq.nlargest(n, lengths, key=lambda x: x[1])))[0])

    with open(src) as f, wf_open(tar) as wf:
        for i, line in tqdm(enumerate(f), "selecting %s most active students from %s to %s" % (n, src, tar)):
            if i not in selected_idx:
                continue
            print(line, end='', file=wf)
